@@ -1,4 +1,4 @@
-from numba import jit, vectorize, float32, float64, cfunc, njit, prange, get_num_threads
+from numba import jit, vectorize, float32, float64, cfunc, njit, prange, get_num_threads, set_num_threads
 import numpy as np
 from scipy.special import comb
 from scipy.interpolate import interp2d, RectBivariateSpline
@@ -370,50 +370,48 @@ def WeightedGridInterp3D(f, wt, x, h, center, size, res=100, box_size=-1):
     wt - (N,) array of weights
     x - (N,3) array of particle positions
     h - (N,) array of particle smoothing lengths
-    center - (2,) array containing the coorindates of the center of the map
+    center - (3,) array containing the coorindates of the center of the map
     size - side-length of the map
     res - resolution of the grid
     """
     dx = size/(res-1)
 
-    x3d = x - center + size/2 - dx/2 # + dx/2 # coordinates in the grid frame, such that the origin is at the corner of the grid
+    x3d = x - center # coordinates in the grid frame, such that the origin is at the corner of the grid 
+    gridcoords = np.linspace(dx/2-size/2, size/2-dx/2, res)
     
     grid = np.zeros((res,res,res))
     gridwt = np.zeros_like(grid)
-    
+
+    x_to_grid_idx = np.empty(3)
     N = len(x)
     for i in range(N):
         xs = x3d[i]        
         hs = h[i]        
         hinv = 1/hs
-#        if True:
-        if box_size < 0:
-            gxmin = max(int((xs[0] - hs)/dx+1),0)
-            gxmax = min(int((xs[0] + hs)/dx),res-1)
-            gymin = max(int((xs[1] - hs)/dx+1), 0)
-            gymax = min(int((xs[1] + hs)/dx),res-1)
-            gzmin = max(int((xs[2] - hs)/dx+1), 0)
-            gzmax = min(int((xs[2] + hs)/dx),res-1)
-        else:
-            gxmin = int((xs[0] - hs)/dx+1)
-            gxmax = int((xs[0] + hs)/dx)
-            gymin = int((xs[1] - hs)/dx+1)
-            gymax = int((xs[1] + hs)/dx)
-            gzmin = int((xs[2] - hs)/dx+1)
-            gzmax = int((xs[2] + hs)/dx)
+        for k in range(3): x_to_grid_idx[k] = (xs[k] + size/2)/dx - 0.5
+        hs_dx = hs/dx
+        gxmin = max(int(x_to_grid_idx[0]-hs_dx+1),0)
+        gxmax = min(int(x_to_grid_idx[0]+hs_dx),res-1)
+        gymin = max(int(x_to_grid_idx[1]-hs_dx+1),0)
+        gymax = min(int(x_to_grid_idx[1]+hs_dx),res-1)
+        gzmin = max(int(x_to_grid_idx[2]-hs_dx+1),0)
+        gzmax = min(int(x_to_grid_idx[2]+hs_dx),res-1)
 
         # first have to do a prepass to get the weight
         kval = np.empty(int(2*hs/dx+1)**3) # save kernel values so don't have to recompute
         total_wt = 0
         j = 0
         for gx in range(gxmin, gxmax+1):            
-            delta_x_Sqr = xs[0] - gx*dx
+#            delta_x_Sqr = xs[0] - gx*dx
+            delta_x_Sqr = xs[0] - gridcoords[gx]
             delta_x_Sqr *= delta_x_Sqr
             for gy in range(gymin,gymax+1):
-                delta_y_Sqr = xs[1] - gy*dx
+#                delta_y_Sqr = xs[1] - gy*dx
+                delta_y_Sqr = xs[1] - gridcoords[gy]                
                 delta_y_Sqr *= delta_y_Sqr
                 for gz in range(gzmin,gzmax+1):
-                    delta_z_Sqr = xs[2] - gz*dx
+                    delta_z_Sqr = xs[2] - gridcoords[gz]
+#                    delta_z_Sqr = xs[2] - gz*dx
                     delta_z_Sqr *= delta_z_Sqr
                     q = np.sqrt(delta_x_Sqr + delta_y_Sqr + delta_z_Sqr) * hinv
                     if q > 1:
@@ -428,26 +426,35 @@ def WeightedGridInterp3D(f, wt, x, h, center, size, res=100, box_size=-1):
         # OK now do the actual deposition
         j = 0
         for gx in range(gxmin, gxmax+1):            
-#            delta_x_Sqr = xs[0] - gx*dx
-#            delta_x_Sqr *= delta_x_Sqr
             for gy in range(gymin,gymax+1):
-#                delta_y_Sqr = xs[1] - gy*dx
-#                delta_y_Sqr *= delta_y_Sqr
                 for gz in range(gzmin,gzmax+1):
-#                    delta_z_Sqr = xs[2] - gz*dx
-#                    delta_z_Sqr *= delta_z_Sqr
-#                    q = np.sqrt(delta_x_Sqr + delta_y_Sqr + delta_z_Sqr) * hinv
-#                    if q > 1:
-#                        kernel = 0
-#                    elif q <= 0.5:
-#                        kernel = 1 - 6*q*q + 6*q*q*q
-#                    else:
-#                        kernel = 2 * (1-q)*(1-q)*(1-q)
                     kernel = kval[j]; j+=1
-                    total_wt += kernel
-                    grid[gx%res,gy%res,gz%res] += f[i] * kernel * wt[i] / total_wt
-                    gridwt[gx%res,gy%res,gz%res] += kernel*wt[i]/total_wt
-    return grid/gridwt
+                    if total_wt > 0:
+                        grid[gx%res,gy%res,gz%res] += f[i] * kernel * wt[i] / total_wt
+                        gridwt[gx%res,gy%res,gz%res] += kernel*wt[i]/total_wt
+                    
+    result = grid/gridwt
+#    do a loop through the grid to check for nans and address if needed
+    pos = np.zeros(3)
+    for i in range(grid.shape[0]):
+        for j in range(grid.shape[1]):
+            for k in range(grid.shape[2]):
+                if gridwt[i,j,k] == 0: # we have to search for the nearest neighbor and take its value - this is usually a rare edge case
+                    pos[0] = gridcoords[i]
+                    pos[1] = gridcoords[j]
+                    pos[2] = gridcoords[k]
+                    min_dist = 1e100
+                    min_dist_idx = -1
+                    for p in range(x3d.shape[0]): # brute force search lol
+                        dist = 0
+                        for d in range(3):
+                            dist += (x3d[p,d] - pos[d])*(x3d[p,d]-pos[d])
+                        if dist < min_dist:
+                            min_dist = dist
+                            min_dist_idx = p
+                    result[i,j,k] = f[min_dist_idx]
+    return result
+                                
 
 @njit(fastmath=True)
 def GridDensity(f, x, h, center, size, res=100, box_size=-1):
@@ -515,33 +522,111 @@ def GridDensity(f, x, h, center, size, res=100, box_size=-1):
         # OK now do the actual deposition
         j = 0
         for gx in range(gxmin, gxmax+1):          
-#            delta_x_Sqr = xs[0] - gx*dx
-#            delta_x_Sqr *= delta_x_Sqr
             for gy in range(gymin,gymax+1):
-#                delta_y_Sqr = xs[1] - gy*dx
-#                delta_y_Sqr *= delta_y_Sqr
                 for gz in range(gzmin,gzmax+1):
-#                    delta_z_Sqr = xs[2] - gz*dx
-#                    delta_z_Sqr *= delta_z_Sqr
-#                    q = np.sqrt(delta_x_Sqr + delta_y_Sqr + delta_z_Sqr) * hinv
-#                    if q > 1:
-#                        kernel = 0
-#                    elif q <= 0.5:
-#                        kernel = 1 - 6*q*q + 6*q*q*q
-#                    else:
-#                        kernel = 2 * (1-q)*(1-q)*(1-q)
                     kernel = kval[j]; j+=1
-                    total_wt += kernel
-                    grid[gx%res,gy%res,gz%res] += f[i] * kernel / total_wt
+                    if total_wt > 0:
+                        grid[gx%res,gy%res,gz%res] += f[i] * kernel / total_wt
     return grid / (dx*dx*dx)
 
-   
-# @jit
-# def ComputeFaces(ngb, ingb, vol, dweights):
-#     N, Nngb, dim = dweights.shape
-#     result = np.zeros_like(dweights)
-#     for i in range(N):
-#         for j in range(Nngb):
-#             result[i,j] += vol[i] * dweights[i,j]
-#             if ingb[i,j] > -1: result[ngb[i,j],ingb[i,j]] -= vol[i] * dweights[i,j]
-#     return result
+
+@njit#(fastmath=True)
+def GridRadTransfer(lum, m, kappa, x, h, gridres, L, center=0, i0=0):
+    """ Simple radiative transfer solver
+
+    Solves the radiative transfer equation with emission and absorption along a grid of sightlines, in multiple bands
+
+    Parameters
+    ----------
+    lum: array_like
+        shape (N, Nbands) array of particle luminosities in the different bands
+    m: array_like
+        shape (N,) array of particle masses
+    kappa: array_like
+        shape (N,) array of particle opacities (dimensions: length^2 / mass)
+    x: array_like
+        shape (N,3) array of particle positions)
+    h: array_like
+        shape (N,) array containing kernel radii of the particles
+    center: array_like
+        shape (3,) array containing the center coordinates of the image
+    gridres: int
+        image resolution
+    L: float
+        size of the image window in length units
+    i0: array_like, optional
+        shape (Nbands,) or (gridres,greidres,Nbands) array of background intensities
+
+    Returns
+    -------
+    image: array_like
+        shape (res,res) array of integrated intensities, in the units of your luminosity units / length units^2 / sr
+    """
+
+#    don't have parallel working yet - trickier than simple surface density map because the order of extinctions and emissions matters
+#    if ncores = -1:
+#        Nchunks = get_num_threads()
+#    else:
+#        set_num_threads(ncores)
+#        Nchunks = ncores
+
+    x -= center
+    order = x[:,2].argsort() # get order for sorting by distance from observer
+
+    lum, kappa, x, h = lum[order], kappa[order], x[order], h[order]
+
+    Nbands = lum.shape[1]
+
+    image = np.zeros((gridres,gridres,Nbands))
+    image += i0
+    
+    dx = L/(gridres-1)
+    N = len(x)
+
+    lh2 = np.empty(Nbands)
+    k = np.empty(Nbands)
+    for i in range(N):
+        # unpack particle properties ##################
+        xs = x[i] + L/2
+        hs = h[i]
+        for b in range(Nbands): # unpack the brightness and opacity
+            lh2[b] = lum[i,b]/(hs*hs)
+            k[b] = kappa[i,b]
+
+        mh2 = m[i]/hs**2
+
+        # done unpacking particle properties ##########
+
+        gxmin = max(int((xs[0] - hs)/dx+1),0)
+        gxmax = min(int((xs[0] + hs)/dx),gridres-1)
+        gymin = max(int((xs[1] - hs)/dx+1), 0)
+        gymax = min(int((xs[1] + hs)/dx), gridres-1)
+        
+        for gx in range(gxmin, gxmax+1):
+            delta_x_Sqr = xs[0] - gx*dx
+            delta_x_Sqr *= delta_x_Sqr
+            for gy in range(gymin,gymax+1):
+                delta_y_Sqr = xs[1] - gy*dx
+                delta_y_Sqr *= delta_y_Sqr
+                r = delta_x_Sqr + delta_y_Sqr
+                if r > hs*hs: continue
+
+                q = np.sqrt(r) / hs                
+                if q <= 0.5:
+                    kernel = 1 - 6*q*q * (1-q)
+                elif q <= 1.0:
+                    a = 1-q
+                    kernel = 2 * a * a * a
+                else:
+                    continue
+                kernel *= 1.8189136353359467
+                
+                for b in range(Nbands):
+                    image[gx,gy,b] += kernel * lh2[b] # emission
+                    tau = k[b] * mh2 # optical depth through the sightline through the particle
+                    if tau < 0.3: # if optically thin use approximation
+                        image[gx,gy,b] *= (1-0.5*tau)/(1+0.5*tau)
+                    else:
+                        image[gx,gy,b] *= np.exp(-tau) #otherwise use the full (more expensive) solution
+                
+    return image / (4*np.pi) 
